# Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
# SPDX-License-Identifier: Apache-2.0
TEST_CERT_DIRECTORY = "../pems/"
TEST_SNI_CERT_DIRECTORY = "../pems/sni/"
TEST_OCSP_DIRECTORY = "../pems/ocsp/"

TRUST_STORE_BUNDLE = TEST_CERT_DIRECTORY + "trust-store/ca-bundle.crt"
TRUST_STORE_TRUSTED_BUNDLE = TEST_CERT_DIRECTORY + "trust-store/ca-bundle.trust.crt"
