#ifndef AWS_IO_FILE_UTILS_H
#define AWS_IO_FILE_UTILS_H

/**
 * Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
 * SPDX-License-Identifier: Apache-2.0.
 */

/* Just shim the code that's in to common, maintain the public interface */
#include <aws/common/file.h>

#endif /* AWS_IO_FILE_UTILS_H */
